<?php

    namespace FederationCLI\InteractiveMode;

    use AsciiTable\Builder;
    use Exception;
    use FederationCLI\InteractiveMode;
    use FederationCLI\Utilities;
    use FederationLib\Enums\FilterOrder;
    use FederationLib\Enums\Filters\ListClientsFilter;
    use FederationLib\Exceptions\DatabaseException;
    use FederationLib\Objects\Client;

    class ClientManager
    {
        /**
         * @param string $input
         * @return void
         */
        public static function processCommand(string $input): void
        {
            $parsed_input = Utilities::parseShellInput($input);

            try
            {
                switch(strtolower($parsed_input['command']))
                {
                    case 'register':
                        self::registerClient();
                        break;

                    case 'list':
                        // list [optional: page] [optional: filter] [optional: order] [optional: max_items]
                        self::listClients($parsed_input['args']);
                        break;

                    case 'total':
                        self::totalClients();
                        break;

                    case 'total_pages':
                        self::totalPages($parsed_input['args']);
                        break;

                    case 'get':
                        self::getClient($parsed_input['args']);
                        break;

                    default:
                        print(sprintf('Unknown command: %s', $parsed_input['command']) . PHP_EOL);
                        break;
                }
            }
            catch(Exception $e)
            {
                Utilities::printExceptionStack($e);
            }
        }

        /**
         * Displays the help message for the client manager
         *
         * @return void
         */
        public static function help(): void
        {
            print('Client manager commands:' . PHP_EOL);
            print('  register - registers a new client with the federation' . PHP_EOL);
            print('  list [optional: page (default 1)] [optional: filter (default created_timestamp)] [optional: order (default asc)] [optional: max_items (default 100)] - lists clients' . PHP_EOL);
            print('  total - gets the total number of clients' . PHP_EOL);
            print('  total_pages [optional: max_items (default 100)] - gets the total number of pages of clients' . PHP_EOL);
            print('  get [client uuid] - gets a client by UUID' . PHP_EOL);

        }

        /**
         * Registers a new client with the federation. prints the UUID of the client if successful.
         *
         * @return void
         */
        private static function registerClient(): void
        {
            $client = new Client();

            $client->setName(Utilities::promptInput('Client name (default: Random): '));
            $client->setDescription(Utilities::promptInput('Client description (default: N/A): '));
            $client->setQueryPermission((int)Utilities::parseBoolean(Utilities::promptInput('Query permission (default: 1): ')));
            $client->setUpdatePermission((int)Utilities::parseBoolean(Utilities::promptInput('Update permission (default: 0): ')));


            try
            {
                $client_uuid = InteractiveMode::getFederationLib()->getClientManager()->registerClient($client);
            }
            catch(Exception $e)
            {
                print('Failed to register client: ' . $e->getMessage() . PHP_EOL);
                Utilities::printExceptionStack($e);
                return;
            }

            print(sprintf('Client registered successfully, UUID: %s', $client_uuid) . PHP_EOL);
        }

        /**
         * @param array $args
         * @return void
         * @throws DatabaseException
         * @throws \Doctrine\DBAL\Exception
         * @throws \RedisException
         */
        private static function listClients(array $args): void
        {
            $page = $args[0] ?? 1;
            $filter = $args[1] ?? ListClientsFilter::CREATED_TIMESTAMP;
            $order = $args[2] ?? FilterOrder::ASCENDING;
            $max_items = $args[3] ?? 100;

            $clients = InteractiveMode::getFederationLib()->getClientManager()->listClients($page, $filter, $order, $max_items);

            if(count($clients) === 0)
            {
                print('No clients found' . PHP_EOL);
            }
            else
            {
                $builder = new Builder();
                foreach($clients as $client)
                {
                    $builder->addRow($client->toArray());
                }
                $builder->setTitle(sprintf('Clients (page %d, filter %s, order %s, max items %d)', $page, $filter, $order, $max_items));
                print($builder->renderTable() . PHP_EOL);
            }
        }

        private static function getClient(mixed $args)
        {
            $client_uuid = $args[0] ?? null;

            if(is_null($client_uuid))
            {
                print('Client UUID required' . PHP_EOL);
                return;
            }

            try
            {
                $client = InteractiveMode::getFederationLib()->getClientManager()->getClient($client_uuid);
            }
            catch(Exception $e)
            {
                print('Failed to get client: ' . $e->getMessage() . PHP_EOL);
                Utilities::printExceptionStack($e);
                return;
            }

            foreach($client->toArray() as $key => $value)
            {
                print match ($key)
                {
                    'id' => (sprintf(' UUID: %s', $value) . PHP_EOL),
                    'enabled' => (sprintf(' Enabled: %s', (Utilities::parseBoolean($value) ? 'Yes' : 'No')) . PHP_EOL),
                    'name' => (sprintf(' Name: %s', $value ?? 'N/A') . PHP_EOL),
                    'description' => (sprintf(' Description: %s', $value ?? 'N/A') . PHP_EOL),
                    'secret_totp' => (sprintf(' Secret TOTP: %s', $value ?? 'N/A') . PHP_EOL),
                    'query_permission' => (sprintf(' Query permission Level: %s', $value) . PHP_EOL),
                    'update_permission' => (sprintf(' Update permission Level: %s', $value) . PHP_EOL),
                    'flags' => (sprintf(' Flags: %s', $value) . PHP_EOL),
                    'created_timestamp' => (sprintf(' Created: %s', date('Y-m-d H:i:s', $value)) . PHP_EOL),
                    'updated_timestamp' => (sprintf(' Updated: %s', date('Y-m-d H:i:s', $value)) . PHP_EOL),
                    'seen_timestamp' => (sprintf(' Last seen: %s', date('Y-m-d H:i:s', $value)) . PHP_EOL),
                    default => (sprintf(' %s: %s', $key, $value) . PHP_EOL),
                };
            }

        }

        private static function totalClients()
        {
            print(sprintf('Total clients: %d', InteractiveMode::getFederationLib()->getClientManager()->getTotalClients()) . PHP_EOL);
        }

        private static function totalPages(mixed $args)
        {
            $max_items = $args[0] ?? 100;

            print(sprintf('Total pages: %d', InteractiveMode::getFederationLib()->getClientManager()->getTotalPages($max_items)) . PHP_EOL);
        }
    }