<?php

    namespace FederationLib\Managers;

    use Exception;
    use FederationLib\Classes\Configuration;
    use FederationLib\Classes\Database;
    use FederationLib\Classes\Redis;
    use FederationLib\Classes\Utilities;
    use FederationLib\Enums\DatabaseTables;
    use FederationLib\Enums\EventPriority;
    use FederationLib\Enums\FilterOrder;
    use FederationLib\Enums\Filters\ListClientsFilter;
    use FederationLib\Enums\Standard\EventCode;
    use FederationLib\Exceptions\DatabaseException;
    use FederationLib\Exceptions\Standard\ClientNotFoundException;
    use FederationLib\FederationLib;
    use FederationLib\Objects\Client;
    use LogLib\Log;
    use Symfony\Component\Uid\Uuid;

    class ClientManager
    {
        /**
         * @var FederationLib
         */
        private FederationLib $federationLib;

        /**
         * ClientManager constructor.
         *
         * @param FederationLib $federationLib
         */
        public function __construct(FederationLib $federationLib)
        {
            $this->federationLib = $federationLib;
        }

        /**
         * Registers a client into the database, returns the UUID that was generated for the client.
         *
         * @param Client $client
         * @return string
         * @throws DatabaseException
         */
        public function registerClient(Client $client): string
        {
            $qb = Database::getConnection()->createQueryBuilder();
            $qb->insert(DatabaseTables::CLIENTS);
            $uuid = Uuid::v4()->toRfc4122();

            foreach($client->toArray() as $key => $value)
            {
                switch($key)
                {
                    case 'id':
                        $qb->setValue($key, ':' . $key);
                        $qb->setParameter($key, $uuid);
                        break;

                    case 'name':
                        if($value === null || strlen($value) === 0 || !preg_match('/^[a-zA-Z0-9_\-]+$/', $value ))
                        {
                            $value = Utilities::generateName(4);
                            Log::debug('net.nosial.federationlib', sprintf('generated name for client: %s', $value));
                        }

                        $qb->setValue($key, ':' . $key);
                        $qb->setParameter($key, substr($value, 0, 64));
                        break;

                    case 'description':
                        if($value !== null)
                        {
                            $qb->setValue($key, ':' . $key);
                            $qb->setParameter($key, substr($value, 0, 255));
                        }
                        break;

                    case 'enabled':
                        $qb->setValue($key, ':' . $key);
                        $qb->setParameter($key, $value ? 1 : 0);
                        break;

                    case 'seen_timestamp':
                    case 'updated_timestamp':
                    case 'created_timestamp':
                        $qb->setValue($key, ':' . $key);
                        $qb->setParameter($key, time());
                        break;

                    default:
                        $qb->setValue($key, ':' . $key);
                        $qb->setParameter($key, $value);
                        break;
                }
            }

            try
            {
                $qb->executeStatement();
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to register client: ' . $e->getMessage(), $e);
            }

            $this->federationLib->getEventLogManager()->logEvent(
                EventCode::CLIENT_CREATED, EventPriority::LOW, null,
                sprintf('Registered client with UUID %s', $uuid)
            );

            Log::info('net.nosial.federationlib', sprintf('Registered client with UUID %s', $uuid));
            return $uuid;
        }

        /**
         * Returns an existing client from the database.
         *
         * @param string|Client $uuid
         * @return Client
         * @throws ClientNotFoundException
         * @throws DatabaseException
         */
        public function getClient(string|Client $uuid): Client
        {
            if($uuid instanceof Client)
            {
                $uuid = $uuid->getUuid();
            }

            if(Configuration::isRedisCacheClientObjectsEnabled())
            {
                try
                {
                    if(Redis::getConnection()?->exists(sprintf('Client<%s>', $uuid)))
                    {
                        return Client::fromArray(Redis::getConnection()?->hGetAll(sprintf('Client<%s>', $uuid)));
                    }
                }
                catch(Exception $e)
                {
                    Log::warning('net.nosial.federationlib', sprintf('Failed to get Client from redis: %s', $e->getMessage()));
                }
            }

            $qb = Database::getConnection()->createQueryBuilder();
            $qb->select('*');
            $qb->from(DatabaseTables::CLIENTS);
            $qb->where('uuid = :uuid');
            $qb->setParameter('uuid', $uuid);

            try
            {
                $result = $qb->executeQuery();

                if($result->rowCount() === 0)
                {
                    throw new ClientNotFoundException($uuid);
                }

                $client = Client::fromArray($result->fetchAssociative());
            }
            catch(ClientNotFoundException $e)
            {
                throw $e;
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to get Client: ' . $e->getMessage(), $e);
            }

            if(Configuration::isRedisCacheClientObjectsEnabled())
            {
                try
                {
                    Redis::getConnection()?->hMSet((string)$client, $client->toArray());
                    Redis::getConnection()?->expire((string)$client, Configuration::getRedisCacheClientObjectsTTL());
                }
                catch(Exception $e)
                {
                    Log::warning('net.nosial.federationlib', sprintf('Failed to cache Client in redis: %s', $e->getMessage()));
                }
            }

            return $client;
        }

        /**
         * Updates a client record in the database, if the client does not exist it will be created.
         * This function is cache aware, if the client is cached it will only update the changed values.
         *
         * @param Client $client
         * @return void
         * @throws DatabaseException
         */
        public function updateClient(Client $client): void
        {
            $cached_client = null;

            if(Configuration::isRedisCacheClientObjectsEnabled())
            {
                try
                {
                    if(Redis::getConnection()?->exists(sprintf('Client<%s>', $client->getUuid())))
                    {
                        $cached_client = Client::fromArray(Redis::getConnection()?->hGetAll(sprintf('Client<%s>', $client->getUuid())));
                    }
                }
                catch(Exception $e)
                {
                    Log::warning('net.nosial.federationlib', sprintf('Failed to get Client from redis: %s', $e->getMessage()));
                }
            }

            $qb = Database::getConnection()->createQueryBuilder();
            $qb->update(DatabaseTables::CLIENTS);
            $qb->set('updated_timestamp', ':updated_timestamp');
            $qb->setParameter('updated_timestamp', time());
            $qb->where('uuid = :uuid');
            $qb->setParameter('uuid', $client->getUuid());

            if($cached_client instanceof Client)
            {
                $data = array_diff($client->toArray(), $cached_client->toArray());
            }
            else
            {
                $data = $client->toArray();
            }

            foreach($data as $key => $value)
            {
                switch($key)
                {
                    case 'uuid':
                    case 'created_timestamp':
                    case 'updated_timestamp':
                    case 'seen_timestamp':
                        break;

                    case 'name':
                        if($value === null || strlen($value) === 0 || !preg_match('/^[a-zA-Z0-9_\-]+$/', $value ))
                        {
                            break;
                        }

                        $qb->set($key, ':' . $key);
                        $qb->setParameter($key, substr($value, 0, 64));
                        break;

                    case 'description':
                        if($value !== null)
                        {
                            $qb->set($key, ':' . $key);
                            $qb->setParameter($key, substr($value, 0, 255));
                        }
                        break;

                    case 'enabled':
                        $qb->set($key, ':' . $key);
                        $qb->setParameter($key, $value ? 1 : 0);
                        break;

                    default:
                        $qb->set($key, ':' . $key);
                        $qb->setParameter($key, $value);
                        break;
                }
            }

            try
            {
                $qb->executeStatement();
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to update client: ' . $e->getMessage(), $e);
            }

            if(Configuration::isRedisCacheClientObjectsEnabled())
            {
                // Update the differences in the cache
                if($cached_client instanceof Client)
                {
                    try
                    {
                        Redis::getConnection()?->hMSet((string)$client, array_diff($client->toArray(), $cached_client->toArray()));
                        Redis::getConnection()?->expire((string)$client, Configuration::getRedisCacheClientObjectsTTL());
                    }
                    catch(Exception $e)
                    {
                        Log::warning('net.nosial.federationlib', sprintf('Failed to cache client in redis: %s', $e->getMessage()));
                    }
                }
                else
                {
                    try
                    {
                        Redis::getConnection()?->hMSet((string)$client, $client->toArray());
                        Redis::getConnection()?->expire((string)$client, $client->getUuid(), Configuration::getRedisCacheClientObjectsTTL());
                    }
                    catch(Exception $e)
                    {
                        Log::warning('net.nosial.federationlib', sprintf('Failed to cache Client in redis: %s', $e->getMessage()));
                    }
                }
            }
        }

        /**
         * Updates a client's last seen timestamp.
         *
         * @param string|Client $uuid
         * @return void
         * @throws DatabaseException
         * @noinspection PhpUnused
         */
        public function updateLastSeen(string|Client $uuid): void
        {
            if($uuid instanceof Client)
            {
                $uuid = $uuid->getUuid();
            }

            $timestamp = time();

            $qb = Database::getConnection()->createQueryBuilder();
            $qb->update(DatabaseTables::CLIENTS);
            $qb->set('seen_timestamp', ':seen_timestamp');
            $qb->setParameter('seen_timestamp', $timestamp);
            $qb->where('uuid = :uuid');
            $qb->setParameter('uuid', $uuid);

            try
            {
                $qb->executeStatement();
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to update last seen timestamp: ' . $e->getMessage(), $e);
            }

            // Update the 'seen_timestamp' only in the hash table in redis
            if(Configuration::isRedisCacheClientObjectsEnabled())
            {
                try
                {
                    Redis::getConnection()?->hSet(sprintf('Client<%s>', $uuid), 'seen_timestamp', $timestamp);
                }
                catch(Exception $e)
                {
                    Log::warning('net.nosial.federationlib', sprintf('Failed to update last seen timestamp in redis: %s', $e->getMessage()));
                }
            }
        }

        /**
         * Returns an array of clients based on the filter, order, and page.
         *
         * @param int $page
         * @param string $filter
         * @param string $order
         * @param int $max_items
         * @return array
         * @throws DatabaseException
         */
        public function listClients(int $page, string $filter=ListClientsFilter::CREATED_TIMESTAMP, string $order=FilterOrder::ASCENDING, int $max_items=100): array
        {
            $qb = Database::getConnection()->createQueryBuilder();
            $qb->select(
                'uuid', 'enabled', 'name', 'description', 'secret_totp', 'query_permission', 'update_permission',
                'flags', 'created_timestamp', 'updated_timestamp', 'seen_timestamp'
            );
            $qb->from(DatabaseTables::CLIENTS);
            $qb->setFirstResult(($page - 1) * $max_items);
            $qb->setMaxResults($max_items);

            if($order !== FilterOrder::ASCENDING && $order !== FilterOrder::DESCENDING)
            {
                throw new DatabaseException('Invalid order: ' . $order);
            }

            switch($filter)
            {
                case ListClientsFilter::CREATED_TIMESTAMP:
                    $qb->orderBy('created_timestamp', strtoupper($order));
                    break;

                case ListClientsFilter::UPDATED_TIMESTAMP:
                    $qb->orderBy('updated_timestamp', strtoupper($order));
                    break;

                case ListClientsFilter::SEEN_TIMESTAMP:
                    $qb->orderBy('seen_timestamp', strtoupper($order));
                    break;

                case ListClientsFilter::ENABLED:
                    $qb->orderBy('enabled', strtoupper($order));
                    break;

                default:
                    throw new DatabaseException('Invalid filter: ' . $filter);
            }

            try
            {
                $result = $qb->executeQuery();
                $clients = [];

                while($row = $result->fetchAssociative())
                {
                    $clients[] = Client::fromArray($row);
                }
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to list clients: ' . $e->getMessage(), $e->getCode(), $e);
            }

            unset($client);
            return $clients;
        }

        /**
         * Returns the total number of clients.
         *
         * @return int
         * @throws DatabaseException
         */
        public function getTotalClients(): int
        {
            $qb = Database::getConnection()->createQueryBuilder();
            $qb->select('COUNT(uuid)');
            $qb->from(DatabaseTables::CLIENTS);

            try
            {
                $result = $qb->executeQuery();
                $row = $result->fetchAssociative();
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to get total clients: ' . $e->getMessage(), $e->getCode(), $e);
            }

            return (int)$row['COUNT(uuid)'];
        }

        /**
         * Returns teh total number of pages.
         *
         * @param int $max_items
         * @return int
         * @throws DatabaseException
         */
        public function getTotalPages(int $max_items=100): int
        {
            return (int)ceil($this->getTotalClients() / $max_items);
        }

        /**
         * Deletes an existing client from the database.
         *
         * @param string|Client $uuid
         * @return void
         * @throws DatabaseException
         */
        public function deleteClient(string|Client $uuid): void
        {
            if($uuid instanceof Client)
            {
                $uuid = $uuid->getUuid();
            }

            $qb = Database::getConnection()->createQueryBuilder();
            $qb->delete(DatabaseTables::CLIENTS);
            $qb->where('uuid = :uuid');
            $qb->setParameter('uuid', $uuid);

            try
            {
                $qb->executeStatement();
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to delete client: ' . $e->getMessage(), $e->getCode(), $e);
            }

            // Invalidate the cache
            if(Configuration::isRedisCacheClientObjectsEnabled())
            {
                try
                {
                    Redis::getConnection()?->del(sprintf('Client<%s>', $uuid));
                }
                catch(Exception $e)
                {
                    Log::warning('net.nosial.federationlib', sprintf('Failed to invalidate client cache in redis: %s', $e->getMessage()));
                }
            }
        }
    }